#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <thread>

#include <dune/solvers/norms/energynorm.hh>
#include <dune/solvers/iterationsteps/multigridstep.hh>
#include <dune/solvers/iterationsteps/cgstep.hh>
#include <dune/solvers/solvers/loopsolver.hh>

#include "../spatial-solving/preconditioners/multilevelpatchpreconditioner.hh"

#include <dune/tectonic/utils/reductionfactors.hh>

#include "adaptivetimestepper.hh"

const unsigned int N_THREADS = std::thread::hardware_concurrency();

template<class IterationStepType, class NormType, class ReductionFactorContainer>
Dune::Solvers::Criterion reductionFactorCriterion(
      IterationStepType& iterationStep,
      const NormType& norm,
      ReductionFactorContainer& reductionFactors)
{
  double normOfOldCorrection = 1;
  auto lastIterate = std::make_shared<typename IterationStepType::Vector>(*iterationStep.getIterate());

  return Dune::Solvers::Criterion(
      [&, lastIterate, normOfOldCorrection] () mutable {
        double normOfCorrection = norm.diff(*lastIterate, *iterationStep.getIterate());
        double convRate = (normOfOldCorrection > 0) ? normOfCorrection / normOfOldCorrection : 0.0;

        if (convRate>1.0)
            std::cout << "Solver convergence rate of " << convRate << std::endl;

        normOfOldCorrection = normOfCorrection;
        *lastIterate = *iterationStep.getIterate();

        reductionFactors.push_back(convRate);
        return std::make_tuple(convRate < 0, Dune::formatString(" % '.5f", convRate));
      },
      " reductionFactor");
}


template<class IterationStepType, class Functional, class ReductionFactorContainer>
Dune::Solvers::Criterion energyCriterion(
      const IterationStepType& iterationStep,
      const Functional& f,
      ReductionFactorContainer& reductionFactors)
{
  double normOfOldCorrection = 1;
  auto lastIterate = std::make_shared<typename IterationStepType::Vector>(*iterationStep.getIterate());

  return Dune::Solvers::Criterion(
      [&, lastIterate, normOfOldCorrection] () mutable {
        double normOfCorrection = std::abs(f(*lastIterate) - f(*iterationStep.getIterate())); //norm.diff(*lastIterate, *iterationStep.getIterate());

        double convRate = (normOfOldCorrection != 0.0) ? 1.0 - (normOfCorrection / normOfOldCorrection) : 0.0;

        if (convRate>1.0)
            std::cout << "Solver convergence rate of " << convRate << std::endl;

        normOfOldCorrection = normOfCorrection;
        *lastIterate = *iterationStep.getIterate();

        reductionFactors.push_back(convRate);
        return std::make_tuple(convRate < 0, Dune::formatString(" % '.5f", convRate));
      },
      " reductionFactor");
}

template <class ReductionFactorContainer>
void updateReductionFactors(ReductionFactorContainer& reductionFactors) {
    const size_t s = reductionFactors.size();

    //print(reductionFactors, "reduction factors: ");

    if (s>allReductionFactors.size()) {
        allReductionFactors.resize(s);
    }

    for (size_t i=0; i<reductionFactors.size(); i++) {
        allReductionFactors[i].push_back(reductionFactors[i]);
    }

    reductionFactors.clear();
}

void IterationRegister::registerCount(FixedPointIterationCounter count) {
  totalCount += count;
}

void IterationRegister::registerFinalCount(FixedPointIterationCounter count) {
  finalCount = count;
}

void IterationRegister::reset() {
  totalCount = FixedPointIterationCounter();
  finalCount = FixedPointIterationCounter();
}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::AdaptiveTimeStepper(
        Dune::ParameterTree const &parset,
        ContactNetwork& contactNetwork,
        const IgnoreVector& ignoreNodes,
        GlobalFriction& globalFriction,
        const std::vector<const BitVector*>& bodywiseNonmortarBoundaries,
        Updaters &current,
        double relativeTime,
        double relativeTau,
        ExternalForces& externalForces,
        const ErrorNorms& errorNorms,
        std::function<bool(Updaters &, Updaters &)> mustRefine)
    : relativeTime_(relativeTime),
      relativeTau_(relativeTau),
      finalTime_(parset.get<double>("problem.finalTime")),
      parset_(parset),
      contactNetwork_(contactNetwork),
      ignoreNodes_(ignoreNodes),
      globalFriction_(globalFriction),
      bodywiseNonmortarBoundaries_(bodywiseNonmortarBoundaries),
      current_(current),
      R1_(),
      externalForces_(externalForces),
      mustRefine_(mustRefine),
      errorNorms_(errorNorms) {}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
bool AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::reachedEnd() {
  return relativeTime_ >= 1.0;
}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
auto AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::makeLinearSolver() const {
    // make linear solver for linear correction in TNNMGStep
    using Norm =  EnergyNorm<Matrix, Vector>;
    using Preconditioner = MultilevelPatchPreconditioner<ContactNetwork, Matrix, Vector>;
    using LinearSolver = typename Dune::Solvers::LoopSolver<Vector>;

    const auto& preconditionerParset = parset_.sub("solver.tnnmg.linear.preconditioner");

    Dune::BitSetVector<1> activeLevels(contactNetwork_.nLevels(), true);
    Preconditioner preconditioner(preconditionerParset, contactNetwork_, activeLevels);
    preconditioner.setPatchDepth(preconditionerParset.template get<size_t>("patchDepth"));
    preconditioner.build();

    auto cgStep = std::make_shared<Dune::Solvers::CGStep<Matrix, Vector>>();
    cgStep->setPreconditioner(preconditioner);

    Norm norm(*cgStep);

    return std::make_shared<LinearSolver>(cgStep, parset_.template get<int>("solver.tnnmg.main.multi"), parset_.template get<double>("solver.tnnmg.linear.tolerance"), norm, Solver::QUIET);
}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
IterationRegister AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::advance() {
  /*
    |     C     | We check here if making the step R1 of size tau is a
    |  R1 | R2  | good idea. To check if we can coarsen, we compare
    |F1|F2|     | the result of (R1+R2) with C, i.e. two steps of size
                  tau with one of size 2*tau. To check if we need to
    refine, we compare the result of (F1+F2) with R1, i.e. two steps
    of size tau/2 with one of size tau. The method makes multiple
    coarsening/refining attempts, with coarsening coming first. */

  //std::cout << "AdaptiveTimeStepper::advance()" << std::endl;

  // patch preconditioner only needs to be computed once per advance()
  // make linear solver for linear correction in TNNMGStep
  using Norm =  EnergyNorm<Matrix, Vector>;
  using Preconditioner = MultilevelPatchPreconditioner<ContactNetwork, Matrix, Vector>;
  using LinearSolver = typename Dune::Solvers::LoopSolver<Vector>;

  /*const auto& preconditionerParset = parset_.sub("solver.tnnmg.preconditioner");

  Dune::BitSetVector<1> activeLevels(contactNetwork_.nLevels(), true);
  Preconditioner preconditioner(preconditionerParset, contactNetwork_, activeLevels);
  preconditioner.setPatchDepth(preconditionerParset.template get<size_t>("patchDepth"));
  preconditioner.build();

  auto cgStep = std::make_shared<Dune::Solvers::CGStep<Matrix, Vector>>();
  cgStep->setPreconditioner(preconditioner);

  Norm norm(*cgStep);

  auto linearSolver = std::make_shared<LinearSolver>(cgStep, parset_.template get<int>("solver.tnnmg.main.multi"), parset_.template get<double>("solver.tnnmg.preconditioner.basesolver.tolerance"), norm, Solver::QUIET);
  */

  // set multigrid solver
  auto smoother = TruncatedBlockGSStep<Matrix, Vector>();

  using TransferOperator = NBodyContactTransfer<ContactNetwork, Vector>;
  using TransferOperators = std::vector<std::shared_ptr<TransferOperator>>;

  TransferOperators transfer(contactNetwork_.nLevels()-1);
  for (size_t i=0; i<transfer.size(); i++) {
      transfer[i] = std::make_shared<TransferOperator>();
      transfer[i]->setup(contactNetwork_, i, i+1);
  }

  // Remove any recompute filed so that initially the full transferoperator is assembled
  for (size_t i=0; i<transfer.size(); i++)
      std::dynamic_pointer_cast<TruncatedMGTransfer<Vector> >(transfer[i])->setRecomputeBitField(nullptr);

  auto linearMultigridStep = std::make_shared<Dune::Solvers::MultigridStep<Matrix, Vector> >();
  linearMultigridStep->setMGType(1, 3, 3);
  linearMultigridStep->setSmoother(smoother);
  linearMultigridStep->setTransferOperators(transfer);

  Norm norm(*linearMultigridStep);

  auto linearSolver = std::make_shared<LinearSolver>(linearMultigridStep, parset_.template get<int>("solver.tnnmg.main.multi"), parset_.template get<double>("solver.tnnmg.preconditioner.basesolver.tolerance"), norm, Solver::QUIET);

  Vector x;
  x.resize(contactNetwork_.nBodyAssembler().totalHasObstacle_.size());
  x = 0;
  dynamic_cast<Dune::Solvers::IterationStep<Vector>*>(linearMultigridStep.get())->setProblem(x);
  //dynamic_cast<Dune::Solvers::IterationStep<Vector>*>(cgStep.get())->setProblem(x);

  const auto& currentNBodyAssembler = contactNetwork_.nBodyAssembler();

   std::vector<double> reductionFactors;
   linearSolver->addCriterion(reductionFactorCriterion(*linearMultigridStep, norm, reductionFactors));
   //linearSolver->addCriterion(reductionFactorCriterion(*cgStep, norm, reductionFactors));

  auto refine = [&](UpdatersWithCount& F1, UpdatersWithCount& F2){
      setDeformation(current_);
      F1 = step(current_, currentNBodyAssembler, linearSolver, relativeTime_, relativeTau_ / 2.0);
      updateReductionFactors(reductionFactors);
      std::cout << "AdaptiveTimeStepper F1 computed!" << std::endl << std::endl;

      setDeformation(F1.updaters);
      auto&& nBodyAssembler = step(currentNBodyAssembler);
      F2 = step(F1.updaters, nBodyAssembler, linearSolver, relativeTime_ + relativeTau_ / 2.0,
                relativeTau_ / 2.0);
      updateReductionFactors(reductionFactors);
  };

  auto coarsen = [&](UpdatersWithCount& C, UpdatersWithCount& R2){
      setDeformation(current_);
      C = step(current_, currentNBodyAssembler, linearSolver, relativeTime_, 2 * relativeTau_).get();
      updateReductionFactors(reductionFactors);
      std::cout << "AdaptiveTimeStepper C computed!" << std::endl << std::endl;

      /*using ScalarVector = typename Updaters::StateUpdater::ScalarVector;
      std::vector<ScalarVector> cAlpha(contactNetwork_.nBodies());
      C.updaters.state_->extractAlpha(cAlpha);
      print(cAlpha, "cAlpha: ");*/

      setDeformation(R1_.updaters);
      auto&& nBodyAssembler = step(currentNBodyAssembler);
      R2 = step(R1_.updaters, nBodyAssembler, linearSolver, relativeTime_ + relativeTau_, relativeTau_).get();
      updateReductionFactors(reductionFactors);
      std::cout << "AdaptiveTimeStepper R2 computed!" << std::endl << std::endl;
  };

  auto canCoarsen = [&]{
    std::cout << N_THREADS << " concurrent threads are supported." << std::endl;

    if (R1_.updaters == Updaters()) {
      //setDeformation(current_);
      R1_ = step(current_, currentNBodyAssembler, linearSolver, relativeTime_, relativeTau_);
      updateReductionFactors(reductionFactors);
    }

    if (N_THREADS<3) {

    } else {

    }

    return ;
  };

  //std::cout << "AdaptiveTimeStepper Step 1" << std::endl;

  size_t coarseningCount = 0;
  size_t refineCount = 0;

  bool didCoarsen = false;
  iterationRegister_.reset();
  UpdatersWithCount R2;
  UpdatersWithCount C;

  while (relativeTime_ + relativeTau_ <= 1.0) {
    std::cout << "tau: " << relativeTau_ << std::endl;

    coarsen(C, R2);

    /*std::vector<ScalarVector> rAlpha(contactNetwork_.nBodies());
    R2.updaters.state_->extractAlpha(rAlpha);
    print(rAlpha, "rAlpha: ");*/

    if (mustRefine_(C.updaters, R2.updaters))
      break;

    didCoarsen = true;
    relativeTau_ *= 2;
    R1_ = C;

    coarseningCount++;
  }

  UpdatersWithCount F1;
  UpdatersWithCount F2;
  if (!didCoarsen) {
    while (true) {
      refine(F1, F2);

      if (!mustRefine_(R1_.updaters, F2.updaters)) {
        std::cout << "Sufficiently refined!" << std::endl;
        break;
      }

      relativeTau_ /= 2.0;
      R1_ = F1;
      R2 = F2;

      refineCount++;
    }
  }

  std::cout << "AdaptiveTimeStepper::advance() ...";

  iterationRegister_.registerFinalCount(R1_.count);
  relativeTime_ += relativeTau_;
  current_ = R1_.updaters;

  //UpdatersWithCount emptyR1;
  //R1_ = emptyR1;
  R1_ = R2;

  std::cout << " done with coarseningCount: " << coarseningCount << " and refineCount: " << refineCount << std::endl;

  return iterationRegister_;
}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
void AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::setDeformation(const Updaters& updaters) {
    std::vector<Vector> u;
    updaters.rate_->extractDisplacement(u);

    for (size_t i=0; i<contactNetwork_.nBodies(); i++) {
        contactNetwork_.body(i)->setDeformation(u[i]);
    }

    // note: levelContactNetworks are not up-to-date; build() has to be called in order to update;
    // unnecessary for standard multigrid as linear solver, might have to be called for patch preconditioner
}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
typename AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::NBodyAssembler
AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::step(const NBodyAssembler& oldNBodyAssembler) const {
    NBodyAssembler nBodyAssembler = oldNBodyAssembler;

    nBodyAssembler.assembleTransferOperator();
    nBodyAssembler.assembleObstacle();

    return nBodyAssembler;
}

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
template <class LinearSolver>
typename AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::UpdatersWithCount
AdaptiveTimeStepper<Factory, ContactNetwork, Updaters, ErrorNorms>::step(
    const Updaters& oldUpdaters, const NBodyAssembler& nBodyAssembler, std::shared_ptr<LinearSolver>& linearSolver, double rTime, double rTau) {

  auto doStep = [&]{
    UpdatersWithCount newUpdatersAndCount = {oldUpdaters.clone(), {}};

    MyCoupledTimeStepper coupledTimeStepper(finalTime_, parset_, nBodyAssembler,
                                          ignoreNodes_, globalFriction_, bodywiseNonmortarBoundaries_,
                                          newUpdatersAndCount.updaters, errorNorms_, externalForces_);

    newUpdatersAndCount.count = coupledTimeStepper.step(linearSolver, rTime, rTau);
    iterationRegister_.registerCount(newUpdatersAndCount.count);

    return newUpdatersAndCount;
  };

  return doStep();
  /*using ScalarVector = typename Updaters::StateUpdater::ScalarVector;
  std::vector<ScalarVector> alpha(contactNetwork_.nBodies());
  newUpdatersAndCount.updaters.state_->extractAlpha(alpha);
  print(alpha, "step alpha: ");
  */

}

#include "adaptivetimestepper_tmpl.cc"
